"""Tests the dataloader."""

import itertools
import random
from typing import Iterator

import numpy as np
import pytest

from dpshdl.collate import collate
from dpshdl.dataloader import Dataloader
from dpshdl.dataset import ChunkedDataset, Dataset


class DummyDataset(Dataset[int, np.ndarray]):
    def start(self) -> None:
        pass

    def next(self) -> int:
        return random.randint(0, 5)

    def collate(self, items: list[int]) -> np.ndarray:
        return collate(items)


@pytest.mark.parametrize("num_workers", [0, 1, 2])
@pytest.mark.parametrize("batch_size", [1, 4])
def test_dataloader(num_workers: int, batch_size: int) -> None:
    ds = DummyDataset()
    with Dataloader(ds, num_workers=num_workers, batch_size=batch_size) as loader:
        for sample in itertools.islice(loader, 10):
            assert sample.shape == (batch_size,)


class DummyChunkedDataset(ChunkedDataset[int, np.ndarray]):
    def __init__(self, value: int) -> None:
        super().__init__()

        self.value = value

    def next_chunk(self) -> Iterator[int]:
        for i in range(self.value):
            yield i

    def collate(self, items: list[int]) -> np.ndarray:
        return collate(items)


@pytest.mark.parametrize("num_workers", [1, 2])
def test_chunked_dataloader(num_workers: int) -> None:
    ds = DummyChunkedDataset(5)
    with Dataloader(ds, num_workers=num_workers, batch_size=2) as loader:
        assert all(t.shape == (2,) for t in itertools.islice(loader, 10))


if __name__ == "__main__":
    # python -m tests.test_dataloader
    test_dataloader(0, 1)



"""Prefetcher tests."""

import itertools
import random

import numpy as np
import pytest

from dpshdl.collate import collate
from dpshdl.dataloader import Dataloader
from dpshdl.dataset import Dataset
from dpshdl.prefetcher import Prefetcher


class DummyDataset(Dataset[int, np.ndarray]):
    def start(self) -> None:
        pass

    def next(self) -> int:
        return random.randint(0, 5)

    def collate(self, items: list[int]) -> np.ndarray:
        return collate(items)


def to_device_fn(sample: np.ndarray) -> np.ndarray:
    return sample


@pytest.mark.parametrize("num_workers", [0, 1, 2])
@pytest.mark.parametrize("batch_size", [1, 4])
def test_dataloader(num_workers: int, batch_size: int) -> None:
    ds = DummyDataset()
    ld = Dataloader(ds, num_workers=num_workers, batch_size=batch_size)
    with Prefetcher(to_device_fn, ld) as pf:
        for sample in itertools.islice(pf, 10):
            assert sample.shape == (batch_size,)


if __name__ == "__main__":
    # python -m tests.test_prefetcher
    test_dataloader(0, 1)



"""Utility functions for testing datasets and dataloaders."""

import itertools
import logging
import re
import time
from typing import Callable, Iterable, TypeVar

from dpshdl.utils import configure_logging

logger = logging.getLogger(__name__)

T = TypeVar("T")


def print_sample(
    index: int,
    sample: T,
    truncate: int | None = 80,
    replace_whitespace: bool = True,
) -> None:
    """Prints a sample from a dataset.

    Args:
        index: The index of the sample.
        sample: The sample to print.
        truncate: If set, truncates the sample to this length.
        replace_whitespace: If set, replaces all whitespace with a single
            space.
    """
    sample_str = str(sample)
    if replace_whitespace:
        sample_str = re.sub(r"\s+", " ", sample_str)
    if truncate is not None and len(sample_str) > truncate:
        sample_str = sample_str[: truncate - 3] + "..."
    logger.info("Sample %d: %s", index, sample_str)


def print_batch(
    indices: list[int],
    samples: list[T],
    truncate: int | None = 80,
    replace_whitespace: bool = True,
) -> None:
    """Prints a batch of samples.

    Args:
        indices: The indices of the samples.
        samples: The samples to print.
        truncate: If set, truncates the sample to this length.
        replace_whitespace: If set, replaces all whitespace with a single
            space.
    """
    batch_str = str(samples)
    if replace_whitespace:
        batch_str = re.sub(r"\s+", " ", batch_str)
    if truncate is not None and len(batch_str) > truncate:
        batch_str = batch_str[: truncate - 3] + "..."
    logger.info("Samples %d - %d: %s", indices[0], indices[-1], batch_str)


def run_test(
    ds: Iterable[T],
    max_samples: int = 10,
    log_interval: int | None = 1,
    print_fn: Callable[[int, T], None] = print_sample,
    batch_fn: Callable[[list[int], list[T]], None] = print_batch,
    batch_size: int | None = None,
    log_batch_interval: int | None = 1,
) -> None:
    """Defines a function for doing adhoc testing of the dataset.

    Args:
        ds: The dataset to test.
        max_samples: The maximum number of samples to test.
        handle_errors: If set, wraps the dataset in an error handling
            wrapper that will catch and log exceptions.
        log_interval: How often to log a sample. If None, don't log any
            samples.
        print_fn: The function to use for printing samples.
        batch_fn: The function to use for printing batches.
        batch_size: Call the ``batch_fn`` on batches of this size.
        log_batch_interval: Log a batch after this many batches.
    """
    configure_logging()
    start_time = time.time()
    samples: tuple[list[int], list[T]] | None = None if batch_size is None else ([], [])
    batch_id = 0
    for i, sample in enumerate(itertools.islice(ds, max_samples)):
        if log_interval is not None and i % log_interval == 0:
            print_fn(i, sample)
        if samples is not None:
            samples[0].append(i)
            samples[1].append(sample)
            if len(samples[0]) == batch_size:
                batch_id += 1
                if batch_id == log_batch_interval:
                    batch_fn(samples[0], samples[1])
                    batch_id = 0
                samples = [], []
    elapsed_time = time.time() - start_time
    samples_per_second = i / elapsed_time
    logger.info("Tested %d samples in %f seconds (%f samples per second)", i + 1, elapsed_time, samples_per_second)



"""Implements the MNIST dataset."""

import array
import gzip
import logging
import struct
from pathlib import Path
from typing import Literal

import numpy as np

from dpshdl.dataset import TensorDataset
from dpshdl.experiments import FileDownloader
from dpshdl.numpy import one_hot as one_hot_fn

logger = logging.getLogger(__name__)

MnistDtype = Literal["int8", "float32"]


class MNIST(TensorDataset[tuple[np.ndarray, np.ndarray]]):
    def __init__(
        self,
        train: bool,
        root_dir: Path,
        dtype: MnistDtype = "int8",
        one_hot: bool = False,
    ) -> None:
        self.train = train
        self.dtype = dtype
        self.one_hot = one_hot

        base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"
        images_name = "train-images-idx3-ubyte.gz" if train else "t10k-images-idx3-ubyte.gz"
        labels_name = "train-labels-idx1-ubyte.gz" if train else "t10k-labels-idx1-ubyte.gz"

        images_path = FileDownloader(
            base_url + images_name,
            "mnist",
            images_name,
            root_dir=root_dir,
        ).ensure_downloaded()

        labels_path = FileDownloader(
            base_url + labels_name,
            "mnist",
            labels_name,
            root_dir=root_dir,
        ).ensure_downloaded()

        with gzip.open(labels_path, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            labels = np.array(array.array("B", fh.read()), dtype=np.uint8)

        with gzip.open(images_path, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            images = np.array(array.array("B", fh.read()), dtype=np.uint8).reshape(num_data, rows, cols)

        # Process the labels and images.
        self.labels = one_hot_fn(labels, 10) if one_hot else labels
        self.images = self.as_dtype(images)

        super().__init__(self.images, self.labels)

    def as_dtype(self, images: np.ndarray) -> np.ndarray:
        if self.dtype == "int8":
            return images
        elif self.dtype == "float32":
            return images.astype(np.float32) / 255.0
        else:
            raise ValueError(f"Unknown dtype: {self.dtype}")



"""Defines an interface for loading data."""

import bdb
import logging
import random
import re
import sys
import time
from abc import ABC, abstractmethod
from collections import Counter, deque
from dataclasses import dataclass
from typing import Callable, Deque, Generic, Iterator, Sequence, TypeVar, final, overload

import numpy as np

from dpshdl.collate import collate
from dpshdl.numpy import worker_chunk
from dpshdl.testing import print_sample, run_test
from dpshdl.utils import TextBlock, render_text_blocks

logger = logging.getLogger(__name__)

T = TypeVar("T")  # The type of the dataset item.
Ti = TypeVar("Ti")  # Input type for the dataset function.
To = TypeVar("To")  # Output type for the dataset function.
Tc = TypeVar("Tc")  # The type of the collated item.
Tco = TypeVar("Tco")  # The output type for the collate function.
Tarrays = TypeVar("Tarrays", bound=tuple[np.ndarray, ...])


class Dataset(Iterator[T], Generic[T, Tc], ABC):
    """Defines the dataset interface.

    Datasets are analogous to a PyTorch iterable datasets that iterates
    forever. This means that there is no concept of an epoch or dataset size.
    """

    def __init__(self) -> None:
        super().__init__()

        self.worker_id = 0
        self.num_workers = 1

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        """Initializes the dataset worker.

        This method is called once per worker when the dataset is used in a
        dataloader.

        Args:
            worker_id: The ID of the worker.
            num_workers: The number of workers in the worker pool.
            lock: A lock shared by all dataset workers in the dataloader.
        """
        self.worker_id = worker_id
        self.num_workers = num_workers

    @abstractmethod
    def next(self) -> T:
        """Returns the next item in the dataset.

        Returns:
            The next item in the dataset.
        """

    def collate(self, items: list[T]) -> Tc | None:
        """Collates  a list of items into a single item.

        Args:
            items: The items in a batch.

        Returns:
            The collated items.
        """
        return collate(items)

    @final
    def __iter__(self) -> "Dataset[T, Tc]":
        # Don't override this! Use `worker_init` instead.
        return self

    @final
    def __next__(self) -> T:
        # Don't override this! Use `next` instead.
        return self.next()

    def sample(
        self,
        batch_size: int,
        max_errors: int | None = 0,
        to_device_func: Callable[[Tc], Tc] = lambda x: x,
    ) -> Tc | None:
        """Gets a sample from the dataset.

        Args:
            batch_size: The number of samples to draw.
            max_errors: The maximum number of errors to allow before raising
                an error. If None, ignore all errors.
            to_device_func: The function to use for moving the sample to the
                device.

        Returns:
            The collated sample, or None if no samples were drawn.
        """
        samples: list[T] = []
        num_errors = 0
        sample_iter = iter(self)
        while len(samples) < batch_size:
            try:
                samples.append(next(sample_iter))
            except (StopIteration, KeyboardInterrupt):
                break
            except Exception as e:
                num_errors += 1
                if max_errors is not None and num_errors >= max_errors:
                    raise
                else:
                    logger.error("Error %d: %s", num_errors, e)
        collated_sample = self.collate(samples)
        return None if collated_sample is None else to_device_func(collated_sample)

    @overload
    def apply(self, item_fn: Callable[[T], To]) -> "Dataset[To, Tc]": ...

    @overload
    def apply(
        self,
        item_fn: Callable[[T], To],
        collate_fn: Callable[[list[To]], Tco],
    ) -> "Dataset[To, Tco]": ...

    def apply(
        self,
        item_fn: Callable[[T], To],
        collate_fn: Callable[[list[To]], Tco] | None = None,
    ) -> "Dataset":
        """Applies a function transformation to items in the dataset.

        Args:
            item_fn: The function to apply to each element of the dataset.
            collate_fn: The new collate function to use. If not provided, the
                default to the base dataset's collate function.

        Returns:
            The new dataset.
        """
        return DatasetFunction(self, item_fn, collate_fn)

    def test(
        self,
        max_samples: int = 10,
        handle_errors: bool = False,
        log_interval: int | None = 1,
        print_fn: Callable[[int, T], None] = print_sample,
        batch_fn: Callable[[list[int], list[T]], None] | None = None,
        batch_size: int | None = None,
        log_batch_interval: int | None = 1,
    ) -> None:
        """Defines a function for doing adhoc testing of the dataset.

        Args:
            max_samples: The maximum number of samples to test.
            handle_errors: If set, wraps the dataset in an error handling
                wrapper that will catch and log exceptions.
            log_interval: How often to log a sample. If None, don't log any
                samples.
            print_fn: The function to use for printing samples.
            batch_fn: The function to use for printing batches.
            batch_size: Call the ``batch_fn`` on batches of this size.
            log_batch_interval: Log a batch after this many batches.
        """
        ds = (
            ErrorHandlingDataset(
                self,
                flush_every_n_steps=max_samples,
                flush_every_n_seconds=None,
            )
            if handle_errors
            else self
        )

        if batch_fn is None:

            def batch_fn(
                indices: list[int],
                samples: list[T],
                truncate: int | None = 80,
                replace_whitespace: bool = True,
            ) -> None:
                batch = ds.collate(samples)
                batch_str = str(batch)
                if replace_whitespace:
                    batch_str = re.sub(r"\s+", " ", batch_str)
                if truncate is not None and len(batch_str) > truncate:
                    batch_str = batch_str[: truncate - 3] + "..."
                logger.info("Samples %d - %d: %s", indices[0], indices[-1], batch_str)

        run_test(
            ds=ds,
            max_samples=max_samples,
            log_interval=log_interval,
            print_fn=print_fn,
            batch_fn=batch_fn,
            batch_size=batch_size,
            log_batch_interval=log_batch_interval,
        )


class DatasetFunction(Dataset[To, Tco], Generic[Ti, To, Tc, Tco], ABC):
    """Defines a function to apply to samples from the dataset.

    This can be used to apply a functional transformation to dataset items.

    Parameters:
        dataset: The dataset to apply the function to.
        item_fn: The function to apply to each element of the dataset; if not
            provided, then the `apply` method must be implemented.
        collate_fn: The new collate function to use; if not provided, default
            to the base dataset's collate function.
    """

    def __init__(
        self,
        dataset: Dataset[Ti, Tc],
        item_fn: Callable[[Ti], To] | None = None,
        collate_fn: Callable[[list[To]], Tco] | None = None,
    ) -> None:
        super().__init__()

        self.base_dataset = dataset
        self.item_fn = item_fn
        self.collate_fn = collate_fn

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        self.base_dataset.worker_init(worker_id, num_workers)

    def collate(self, items: list[To]) -> Tco | None:
        if self.collate_fn is not None:
            return self.collate_fn(items)
        return self.base_dataset.collate(items)  # type: ignore[arg-type, return-value]

    def apply_function_to_item(self, item: Ti) -> To:
        if self.fn is not None:
            return self.fn(item)
        raise NotImplementedError(
            "Either the `fn` attribute must be provided or the `apply_function_to_item` method must be overridden."
        )

    def next(self) -> To:
        return self.apply_function_to_item(self.base_dataset.next())


class TensorDataset(Dataset[Tarrays, Tarrays], Generic[Tarrays]):
    """Defines a dataset that yields samples from a tensor.

    All provided tensors should have the same shape in the ``dim`` dimension.

    Parameters:
        tensors: The tensors to sample from.
        dim: The dimension to sample from.
        stack_dim: The dimension to stack the tensors on.
    """

    def __init__(self, *tensors: np.ndarray, dim: int = 0, stack_dim: int = 0) -> None:
        super().__init__()

        self.tensors = tensors
        self.dim = dim
        self.stack_dim = stack_dim

        self._worker_tensors = list(tensors)

        # Gets the number of samples.
        self.num_samples = tensors[0].shape[dim]
        self._worker_num_samples = self.num_samples
        if not all(t.shape[dim] == self.num_samples for t in tensors):
            raise ValueError("All tensors must have the same shape in the specified dimension.")

        self.rand = np.random.RandomState(0)

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        self._worker_tensors = [worker_chunk(t, worker_id, num_workers) for t in self.tensors]
        self._worker_num_samples = self._worker_tensors[0].shape[self.dim]
        if not all(t.shape[self.dim] == self._worker_num_samples for t in self._worker_tensors):
            raise ValueError("All tensors must have the same shape in the specified dimension.")

    def next(self) -> Tarrays:
        index = self.rand.randint(0, self._worker_num_samples)
        return tuple(np.take(t, index, axis=0) for t in self._worker_tensors)  # type: ignore[return-value]

    def collate(self, items: list[Tarrays]) -> Tarrays:
        collated_items = (np.stack([item[i] for item in items]) for i in range(len(items[0])))
        return tuple(collated_items)  # type: ignore[return-value]


class ChunkedDataset(Dataset[T, Tc], Generic[T, Tc], ABC):
    """Defines a dataset that yields chunks of samples in a separate thread.

    This dataset is useful for implementing IO bound datasets such as
    datasets that read from disk or download data from the internet. Subclasses
    should implement the `next_chunk` method, which iterates over the next
    chunk of data. For example, a dataset that reads from a spinning disk
    might implement `next_chunk` as follows:

    .. code-block:: python

        def next_chunk(self) -> Iterator[T]:
            path = self.paths[self.current_path]
            self.current_path = (self.current_path + 1) % len(self.paths)
            with open(path, "rb", buffering=BIG_BUFFERING_SIZE) as fh:
                for line in fh:
                    yield line

    In the above example, we use a large buffer size to reduce the number of
    IO operations. Since this happens in a coroutine, we don't need to
    worry about creating a new file pointer for each chunk.

    Parameters:
        max_queue_size: The maximum number of samples to keep in the
            queue at a time.
    """

    def __init__(self, max_queue_size: int = 0) -> None:
        super().__init__()

        self._max_queue_size = max_queue_size
        self._next_chunk_iterator: Iterator[T] | None = None

    @abstractmethod
    def next_chunk(self) -> Iterator[T]:
        """Returns the next chunk of data.

        Returns:
            The next chunk of data.
        """

    def chunked_dataset_iterator(self) -> Iterator[T]:
        while True:
            for sample in self.next_chunk():
                yield sample

    def next(self) -> T:
        if self._next_chunk_iterator is None:
            self._next_chunk_iterator = self.chunked_dataset_iterator()
        try:
            sample = next(self._next_chunk_iterator)
        except Exception:
            self._next_chunk_iterator = None
            raise
        return sample


class RoundRobinDataset(Dataset[T, Tc], Generic[T, Tc]):
    """Defines a dataset that yields samples in round robin fashion.

    Parameters:
        datasets: The datasets to sample from.
    """

    def __init__(
        self,
        datasets: Sequence[Dataset[T, Tc]],
        collate_fn: Callable[[list[T]], Tc | None] = collate,
    ) -> None:
        super().__init__()

        self.datasets = datasets
        self.collate_fn = collate_fn

        self.i = 0

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        for dataset in self.datasets:
            dataset.worker_init(worker_id, num_workers)

    def next(self) -> T:
        next_item = self.datasets[self.i].next()
        self.i = (self.i + 1) % len(self.datasets)
        return next_item

    def collate(self, items: list[T]) -> Tc | None:
        return self.collate_fn(items)


class RandomDataset(Dataset[T, Tc], Generic[T, Tc]):
    """Defines a dataset that randomly samples from a list of datasets.

    Parameters:
        datasets: The datasets to sample from.
    """

    def __init__(
        self,
        datasets: Sequence[Dataset[T, Tc]],
        collate_fn: Callable[[list[T]], Tc | None] = collate,
        stop_on_first: bool = False,
    ) -> None:
        super().__init__()

        self.datasets = datasets
        self.collate_fn = collate_fn
        self.stop_on_first = stop_on_first

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        for dataset in self.datasets:
            dataset.worker_init(worker_id, num_workers)

    def next(self) -> T:
        return random.choice(self.datasets).next()

    def collate(self, items: list[T]) -> Tc | None:
        return self.collate_fn(items)


class InMemoryDataset(Dataset[T, Tc], Generic[T, Tc]):
    """Repeatedly yields from a pool of samples which are stored in-memory.

    Parameters:
        dataset: The dataset to draw samples for the pool from.
        num_samples: The maximum size of the pool.
        yield_in_order: If True, yield the samples in the order they were
            drawn from the original dataset, otherwise yield randomly.
    """

    def __init__(
        self,
        dataset: Dataset[T, Tc],
        num_samples: int,
        yield_in_order: bool = False,
    ) -> None:
        super().__init__()

        self.dataset = dataset
        self.num_samples = num_samples
        self.yield_in_order = yield_in_order

        self.pool: Deque[T] = deque()

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        self.dataset.worker_init(worker_id, num_workers)

    def next(self) -> T:
        if len(self.pool) < self.num_samples:
            self.pool.append(self.dataset.next())
        if self.yield_in_order:
            item = self.pool[0]
            self.pool.rotate(-1)
        else:
            item = self.pool[random.randint(0, len(self.pool) - 1)]
        return item

    def collate(self, items: list[T]) -> Tc | None:
        return self.dataset.collate(items)


def get_loc(num_excs: int = 1) -> str:
    _, _, exc_tb = sys.exc_info()
    if exc_tb is None or (exc_tb := exc_tb.tb_next) is None:
        return "unknown"
    exc_strs: list[str] = []
    for _ in range(num_excs):
        exc_strs += [f"{exc_tb.tb_frame.f_code.co_filename}:{exc_tb.tb_lineno}"]
        if (exc_tb := exc_tb.tb_next) is None:
            break
    return "\n".join(exc_strs)


@dataclass(frozen=True)
class ExceptionSummary:
    title: str
    num_steps: int
    elapsed_time: float
    num_exceptions: int
    top_exception_messages: list[tuple[str, int]]
    top_exception_types: list[tuple[str, int]]
    top_exception_locations: list[tuple[str, int]]
    last_exception: Exception | None
    show_end: int | float | None
    max_lines: int | None

    def __str__(self) -> str:
        blocks: list[list[TextBlock]] = []

        blocks += [
            [
                TextBlock(
                    f"{self.title} ({self.num_steps} steps, {self.elapsed_time:.2f} seconds)",
                    color="red",
                    bold=True,
                    width=60,
                    center=True,
                ),
                TextBlock("Count", color="yellow", bold=False, width=10, center=True),
                TextBlock("Percent", color="yellow", bold=False, width=10, center=True),
            ],
        ]

        def get_header(s: str) -> list[list[TextBlock]]:
            return [
                [
                    TextBlock(s, color="yellow", bold=True, width=60),
                    TextBlock("", width=10),
                    TextBlock("", width=10),
                ],
            ]

        def get_line(ks: str, v: int) -> list[list[TextBlock]]:
            line = [
                TextBlock(ks, width=60, no_sep=True, show_end=self.show_end, max_lines=self.max_lines),
                TextBlock(f"{v}", width=10, no_sep=True),
                TextBlock(f"{v / self.num_steps * 100:.2f}%", width=10, no_sep=True),
            ]
            return [line]

        # Logs unique exception strings.
        blocks += get_header("Exceptions")
        for k, v in self.top_exception_messages:
            blocks += get_line(k, v)

        # Logs the individual exception classes.
        blocks += get_header("Types")
        for k, v in self.top_exception_types:
            blocks += get_line(k, v)

        # Logs by line number.
        blocks += get_header("Locations")
        for k, v in self.top_exception_locations:
            blocks += get_line(k, v)

        # Logs the total number of exceptions.
        blocks += get_header("Total")
        blocks += get_line("Total", self.num_exceptions)

        return render_text_blocks(blocks)


class ExceptionSummaryWriter:
    """Defines a utility class for storing and logging exceptions.

    Parameters:
        title: The title for each summary.
        max_exceptions: The maximum number of unique exceptions to log.
        show_end: How much of the ending of each exception to show.
        max_lines: The maximum number of lines to show of each error message.
    """

    def __init__(
        self,
        title: str,
        max_exceptions: int = 5,
        show_end: int | float | None = 0.5,
        max_lines: int | None = 3,
    ) -> None:
        super().__init__()

        self.title = title
        self.max_exceptions = max_exceptions
        self.show_end = show_end
        self.max_lines = max_lines

        self.exceptions: Counter[str] = Counter()
        self.exception_classes: Counter[str] = Counter()
        self.exception_locs: Counter[str] = Counter()

        self.last_exception: Exception | None = None
        self.num_steps = 0
        self.start_time = time.time()
        self.step_has_error = False
        self.total_exceptions = 0

    @property
    def elapsed_time(self) -> float:
        return time.time() - self.start_time

    def next(self) -> None:
        self.num_steps += 1
        self.step_has_error = False

    def __len__(self) -> int:
        return len(self.exceptions)

    def __bool__(self) -> bool:
        return len(self.exceptions) > 0

    def add_exception(self, exc: Exception, loc: str) -> None:
        self.last_exception = exc
        self.exceptions[f"{exc.__class__.__name__}: {exc}"] += 1
        self.exception_classes[exc.__class__.__name__] += 1
        self.exception_locs[loc] += 1
        if not self.step_has_error:
            self.step_has_error = True
            self.total_exceptions += 1

    def summary(self) -> ExceptionSummary:
        return ExceptionSummary(
            title=self.title,
            num_steps=self.num_steps,
            elapsed_time=self.elapsed_time,
            num_exceptions=self.total_exceptions,
            top_exception_messages=self.exceptions.most_common(self.max_exceptions),
            top_exception_types=self.exception_classes.most_common(self.max_exceptions),
            top_exception_locations=self.exception_locs.most_common(self.max_exceptions),
            last_exception=self.last_exception,
            show_end=self.show_end,
            max_lines=self.max_lines,
        )

    def clear(self) -> None:
        self.exceptions.clear()
        self.exception_classes.clear()
        self.exception_locs.clear()

        self.num_steps = 0
        self.start_time = time.time()
        self.step_has_error = False
        self.total_exceptions = 0

    def __str__(self) -> str:
        return str(self.summary())


class ErrorHandlingDataset(Dataset[T, Tc]):
    """Defines a wrapper for safely handling errors in iterable datasets.

    Parameters:
        dataset: The dataset to wrap.
        sleep_backoff: The initial sleep time after an exception.
        sleep_backoff_power: The power to raise the sleep time by after
            each consecutive exception.
        maximum_exceptions: The maximum number of consecutive exceptions
            to allow before raising an error.
        backoff_after: The number of exceptions to allow before backing
            off (i.e. increasing the sleep time).
        traceback_depth: The number of stack frames to include in the
            exception traceback.
        flush_every_n_steps: Flush the exception summary every N steps.
        flush_every_n_seconds: Flush the exception summary every N seconds.
        log_exceptions_all_workers: If set, log exceptions from all workers.
        log_level: The log level to use for logging exceptions.
    """

    def __init__(
        self,
        dataset: Dataset[T, Tc],
        sleep_backoff: float = 0.1,
        sleep_backoff_power: float = 2.0,
        maximum_exceptions: int = 5,
        backoff_after: int = 5,
        traceback_depth: int = 3,
        flush_every_n_steps: int | None = None,
        flush_every_n_seconds: float | None = 60.0,
        log_exceptions_all_workers: bool = False,
        log_level: int = logging.INFO,
    ) -> None:
        super().__init__()

        self.dataset = dataset
        self.sleep_backoff = sleep_backoff
        self.sleep_backoff_power = sleep_backoff_power
        self.maximum_exceptions = maximum_exceptions
        self.backoff_after = backoff_after
        self.traceback_depth = traceback_depth
        self.flush_every_n_steps = flush_every_n_steps
        self.flush_every_n_seconds = flush_every_n_seconds
        self.log_exceptions_all_workers = log_exceptions_all_workers
        self.log_exceptions = True
        self.log_level = log_level

        self.exc_summary = ExceptionSummaryWriter("Error Summary")
        self.col_exc_summary = ExceptionSummaryWriter("Collate Error Summary")

    def should_flush_summary(self) -> bool:
        if self.flush_every_n_steps is not None and self.exc_summary.num_steps >= self.flush_every_n_steps:
            return True
        if self.flush_every_n_seconds is not None and self.exc_summary.elapsed_time >= self.flush_every_n_seconds:
            return True
        return False

    def should_flush_col_summary(self) -> bool:
        if self.flush_every_n_steps is not None and self.col_exc_summary.num_steps >= self.flush_every_n_steps:
            return True
        if self.flush_every_n_seconds is not None and self.col_exc_summary.elapsed_time >= self.flush_every_n_seconds:
            return True
        return False

    def worker_init(self, worker_id: int, num_workers: int) -> None:
        super().worker_init(worker_id, num_workers)

        self.dataset.worker_init(worker_id, num_workers)
        if worker_id != 0 and not self.log_exceptions_all_workers:
            self.log_exceptions = False

    def next(self) -> T:
        num_exceptions = 0
        backoff_time = self.sleep_backoff
        self.exc_summary.next()

        if self.should_flush_summary():
            if self.log_exceptions and self.exc_summary:
                logger.log(self.log_level, "\n%s", self.exc_summary.summary())
            self.exc_summary.clear()

        while num_exceptions < self.maximum_exceptions:
            try:
                return self.dataset.next()
            except (bdb.BdbQuit, KeyboardInterrupt, StopIteration):
                raise
            except Exception as e:
                self.exc_summary.add_exception(e, get_loc(self.traceback_depth))
            num_exceptions += 1
            if num_exceptions > self.backoff_after:
                logger.error("Encountered %d exceptions, backing off for %f seconds", num_exceptions, backoff_time)
                time.sleep(backoff_time)
                backoff_time *= self.sleep_backoff_power
        raise RuntimeError(f"Reached max exceptions {self.maximum_exceptions}\n{self.exc_summary.summary()}")

    def collate(self, items: list[T]) -> Tc | None:
        self.col_exc_summary.next()

        if self.should_flush_col_summary():
            if self.log_exceptions and self.col_exc_summary:
                logger.log(self.log_level, "\n%s", self.col_exc_summary.summary())
            self.col_exc_summary.clear()

        try:
            return self.dataset.collate(items)
        except (bdb.BdbQuit, KeyboardInterrupt, StopIteration):
            raise
        except Exception as e:
            self.col_exc_summary.add_exception(e, get_loc(self.traceback_depth))
            return None


def test_error_handling_dataset_adhoc() -> None:
    """Tests the error handling dataset."""

    class DummyDataset(Dataset[int, list[int]]):
        def next(self) -> int:
            n = random.random()
            if n < 0.1:
                assert False
            elif n < 0.2:
                raise ValueError("ValueError")
            elif n < 0.3:
                1 / 0
            return random.randint(0, 10)

        def collate(self, items: list[int]) -> list[int]:
            return items

    DummyDataset().test(
        max_samples=100,
        handle_errors=True,
        print_fn=lambda i, sample: logger.info("Sample %d: %d", i, sample),
        batch_size=10,
    )


if __name__ == "__main__":
    # python -m dpshdl.dataset
    test_error_handling_dataset_adhoc()



"""Functions for managing experiments."""

import contextlib
import hashlib
import itertools
import logging
import os
import re
import tempfile
import urllib.error
import urllib.request
import warnings
from pathlib import Path
from typing import Iterator
from urllib.parse import urlparse

import requests

USER_AGENT = "dpshdl"

logger = logging.getLogger(__name__)


def check_md5(file_path: str | Path, hash_str: str | None, chunk_size: int = 2**16) -> bool:
    """Checks the MD5 of the downloaded file.

    Args:
        file_path: Path to the downloaded file.
        hash_str: Expected MD5 of the file; if None, return True.
        chunk_size: Size of the chunks to read from the file.

    Returns:
        True if the MD5 matches, False otherwise.
    """
    if hash_str is None:
        return True

    md5 = hashlib.md5()

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            md5.update(chunk)

    return md5.hexdigest() == hash_str


def check_sha256(file_path: str | Path, hash_str: str | None, chunk_size: int = 2**16) -> bool:
    """Checks the SHA256 of the downloaded file.

    Args:
        file_path: Path to the downloaded file.
        hash_str: Expected SHA256 of the file; if None, return True.
        chunk_size: Size of the chunks to read from the file.

    Returns:
        True if the SHA256 matches, False otherwise.
    """
    if hash_str is None:
        return True

    sha256 = hashlib.sha256()

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            sha256.update(chunk)

    return sha256.hexdigest() == hash_str


class FileDownloader:
    """Provides a simple interface for downloading URLs.

    This class is meant to be subclassed to provide different download
    locations. For example, when downloading pretrained models, use the
    :class:`ModelDownloader` class.

    Typically, you should simply use the :func:`ensure_downloaded` function
    to make sure the file is downloaded to the correct location.

    This is adapted in large part from the reference implementation in the
    Torchvision library.

    Parameters:
        url: The URL to download from.
        dnames: The directory names to download to.
        md5: The expected MD5 of the file.
        sha256: The expected SHA256 of the file.
        is_tmp: Whether to download to a temporary directory.
        recheck_hash: Whether to recheck the hash after downloading.
        max_redirect_hops: The maximum number of redirects to follow.
    """

    def __init__(
        self,
        url: str,
        *dnames: str,
        md5: str | None = None,
        sha256: str | None = None,
        root_dir: Path | None = None,
        is_tmp: bool = False,
        recheck_hash: bool = False,
        max_redirect_hops: int = 3,
    ) -> None:
        super().__init__()

        assert len(dnames) >= 1, "Must provide at least 1 directory name"
        if is_tmp:
            filepath = Path(tempfile.mkdtemp("models"))
        else:
            filepath = Path(root_dir or Path.cwd() / "data")
        for dname in dnames:
            filepath = filepath / dname
        (root := filepath.parent).mkdir(parents=True, exist_ok=True)

        self.url = url
        self.filename = filepath.name
        self.root = root
        self.md5 = md5
        self.sha256 = sha256
        self.recheck_hash = recheck_hash
        self.max_redirect_hops = max_redirect_hops

    @property
    def filepath(self) -> Path:
        return self.root / self.filename

    @property
    def is_downloaded(self) -> bool:
        if not self.filepath.exists():
            return False
        if self.recheck_hash and not self.check_hashes():
            logger.warning("A file was found for %s in %s, but its hashes do not match.", self.url, self.filepath)
            self.filepath.unlink()
            return False
        return True

    def check_hashes(self) -> bool:
        return check_sha256(self.filepath, self.sha256) and check_md5(self.filepath, self.md5)

    def ensure_downloaded(self) -> Path:
        """Ensures the file is downloaded and returns the path to it.

        By default, we only check the hash once when the file is downloaded,
        and we don't bother rechecking unless ``recheck_hash`` is set to True.

        Returns:
            The path to the downloaded file.
        """
        if not self.is_downloaded:
            self.download()
            if not self.check_hashes():
                self.filepath.unlink()
                raise RuntimeError(f"Hashes for {self.filepath} do not match. The corruped file has been deleted.")
        return self.filepath

    def download(self) -> None:
        root = self.root.expanduser()
        root.mkdir(parents=True, exist_ok=True)

        # Expands the redirect chain if needed.
        url = self._get_redirect_url(self.url, max_hops=self.max_redirect_hops)

        # Checks if file is located on Google Drive.
        file_id = self._get_google_drive_file_id(url)
        if file_id is not None:
            return self.download_file_from_google_drive(file_id, root, self.filename)

        # Downloads the file.
        try:
            logger.info("Downloading %s to %s", url, self.filepath)
            self._urlretrieve(url, self.filepath)
        except (urllib.error.URLError, OSError) as e:
            if url[:5] == "https":
                url = url.replace("https:", "http:")
                logger.warning("Download failed. Trying HTTP instead of HTTPS: %s to %s", url, self.filepath)
                self._urlretrieve(url, self.filepath)
            else:
                raise e

    @classmethod
    def _save_response_content(cls, content: Iterator[bytes], destination: Path) -> None:
        with open(destination, "wb") as fh:
            for chunk in content:
                if not chunk:  # Filter out keep-alive new chunks.
                    continue
                fh.write(chunk)

    @classmethod
    def _urlretrieve(cls, url: str, filename: Path, chunk_size: int = 1024 * 32) -> None:
        with urllib.request.urlopen(urllib.request.Request(url, headers={"User-Agent": USER_AGENT})) as response:
            cls._save_response_content(iter(lambda: response.read(chunk_size), b""), filename)

    @classmethod
    def _extract_gdrive_api_response(
        cls,
        response: requests.Response,
        chunk_size: int = 32 * 1024,
    ) -> tuple[str | None, Iterator[bytes]]:
        content = response.iter_content(chunk_size)
        first_chunk = None
        while not first_chunk:  # Filter out keep-alive new chunks.
            first_chunk = next(content)
        content = itertools.chain([first_chunk], content)

        try:
            match = re.search("<title>Google Drive - (?P<api_response>.+?)</title>", first_chunk.decode())
            api_response = match["api_response"] if match is not None else None
        except UnicodeDecodeError:
            api_response = None
        return api_response, content

    @classmethod
    def download_file_from_google_drive(cls, file_id: str, root: Path, filename: str | None = None) -> None:
        root = root.expanduser()
        if not filename:
            filename = file_id
        fpath = root / filename
        root.mkdir(parents=True, exist_ok=True)

        url = "https://drive.google.com/uc"
        params = dict(id=file_id, export="download")
        with requests.Session() as session:
            response = session.get(url, params=params, stream=True)

            for key, value in response.cookies.items():
                if key.startswith("download_warning"):
                    token = value
                    break
            else:
                api_response, content = cls._extract_gdrive_api_response(response)
                token = "t" if api_response == "Virus scan warning" else None

            if token is not None:
                response = session.get(url, params=dict(params, confirm=token), stream=True)
                api_response, content = cls._extract_gdrive_api_response(response)

            if api_response == "Quota exceeded":
                raise RuntimeError(
                    f"The daily quota of the file {filename} is exceeded and it "
                    f"can't be downloaded. This is a limitation of Google Drive "
                    f"and can only be overcome by trying again later."
                )

            cls._save_response_content(content, fpath)

        # In case we deal with an unhandled GDrive API response, the file should be smaller than 10kB with only text.
        if os.stat(fpath).st_size < 10 * 1024:
            with contextlib.suppress(UnicodeDecodeError), open(fpath) as fh:
                text = fh.read()

                # Regular expression to detect HTML. Copied from https://stackoverflow.com/a/70585604
                if re.search(r"</?\s*[a-z-][^>]*\s*>|(&(?:[\w\d]+|#\d+|#x[a-f\d]+);)", text):
                    warnings.warn(
                        f"We detected some HTML elements in the downloaded file. "
                        f"This most likely means that the download triggered an unhandled API response by GDrive. "
                        f"Please report this to torchvision at https://github.com/pytorch/vision/issues including "
                        f"the response:\n\n{text}"
                    )

    @classmethod
    def _get_google_drive_file_id(cls, url: str) -> str | None:
        parts = urlparse(url)
        if re.match(r"(drive|docs)[.]google[.]com", parts.netloc) is None:
            return None
        match = re.match(r"/file/d/(?P<id>[^/]*)", parts.path)
        if match is None:
            return None
        return match.group("id")

    @classmethod
    def _get_redirect_url(cls, url: str, max_hops: int = 3) -> str:
        initial_url = url
        headers = {"Method": "HEAD", "User-Agent": USER_AGENT}
        for _ in range(max_hops + 1):
            with urllib.request.urlopen(urllib.request.Request(url, headers=headers)) as response:
                if response.url == url or response.url is None:
                    return url
                url = response.url
        raise RecursionError(f"Request to {initial_url} exceeded {max_hops} redirects. The last redirect was {url}.")



"""Defines a utility class for pre-loading samples into device memory.

When you are training a model, you usually get a sample from the dataloader,
then need to move it into device memory. This host-to-device transfer can be
slow, so it is beneficial to pre-load the next sample into device memory while
the current sample is being processed.
"""

from types import TracebackType
from typing import Callable, Generic, Iterable, Iterator, TypeVar

from dpshdl.dataloader import Dataloader
from dpshdl.testing import print_sample, run_test

T = TypeVar("T")
Tc_co = TypeVar("Tc_co", covariant=True)
Tp_co = TypeVar("Tp_co", covariant=True)


class Prefetcher(Iterable[Tp_co], Generic[Tc_co, Tp_co]):
    """Helper class for pre-loading samples into device memory."""

    def __init__(self, to_device_func: Callable[[Tc_co], Tp_co], dataloader: Dataloader[T, Tc_co]) -> None:
        super().__init__()

        self.to_device_func = to_device_func
        self.dataloader = dataloader
        self.next_sample: Tp_co | None = None

    def get_sample(self) -> Tp_co:
        return self.to_device_func(next(self.dataloader))

    def __iter__(self) -> Iterator[Tp_co]:
        return self

    def __next__(self) -> Tp_co:
        if self.next_sample is None:
            self.next_sample = self.get_sample()
        sample, self.next_sample = self.next_sample, self.get_sample()
        return sample

    def __enter__(self) -> "Prefetcher[Tc_co, Tp_co]":
        self.dataloader.__enter__()
        return self

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        self.dataloader.__exit__(_t, _e, _tr)

    def test(
        self,
        max_samples: int = 10,
        log_interval: int | None = 1,
        print_fn: Callable[[int, Tp_co], None] = print_sample,
    ) -> None:
        """Defines a function for doing adhoc testing of the prefetcher.

        Args:
            max_samples: The maximum number of samples to test.
            log_interval: How often to log a sample. If None, don't log any
                samples.
            print_fn: The function to use for printing samples.
        """
        with self:
            run_test(
                ds=self,
                max_samples=max_samples,
                log_interval=log_interval,
                print_fn=print_fn,
            )



"""Tests the dataset classes."""

import itertools
from typing import Iterator

import numpy as np

from dpshdl.collate import collate
from dpshdl.dataset import ChunkedDataset, Dataset, RandomDataset, RoundRobinDataset


class DummyDataset(Dataset[int, np.ndarray]):
    def __init__(self, value: int) -> None:
        super().__init__()

        self.value = value

    def next(self) -> int:
        return self.value

    def collate(self, items: list[int]) -> np.ndarray:
        return collate(items)


def test_dataset_simple() -> None:
    ds = DummyDataset(1)
    for sample in itertools.islice(ds, 10):
        assert sample == 1


def test_round_robin_dataset() -> None:
    dss = [DummyDataset(i) for i in range(5)]
    ds = RoundRobinDataset(dss, dss[0].collate)
    assert list(itertools.islice(ds, 10)) == [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]


def test_random_dataset() -> None:
    dss = [DummyDataset(i) for i in range(5)]
    ds = RandomDataset(dss, dss[0].collate)
    assert all(0 <= i < 5 for i in itertools.islice(ds, 10))


def test_apply_function() -> None:
    ds = DummyDataset(5).apply(lambda x: str(x))
    assert list(itertools.islice(ds, 10)) == ["5"] * 10


class DummyChunkedDataset(ChunkedDataset[int, np.ndarray]):
    def __init__(self, value: int) -> None:
        super().__init__()

        self.value = value

    def next_chunk(self) -> Iterator[int]:
        for i in range(self.value):
            yield i

    def collate(self, items: list[int]) -> np.ndarray:
        return collate(items)


def test_chunked_dataset() -> None:
    ds = DummyChunkedDataset(5)
    assert list(itertools.islice(ds, 10)) == [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]


if __name__ == "__main__":
    # python -m tests.test_dataset
    test_round_robin_dataset()



"""Defines some general-purpose utility functions."""

import datetime
import itertools
import logging
import re
import sys
from typing import Literal

RESET_SEQ = "\033[0m"
REG_COLOR_SEQ = "\033[%dm"
BOLD_COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

Color = Literal[
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "grey",
    "light-red",
    "light-green",
    "light-yellow",
    "light-blue",
    "light-magenta",
    "light-cyan",
]

COLOR_INDEX: dict[Color, int] = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "grey": 90,
    "light-red": 91,
    "light-green": 92,
    "light-yellow": 93,
    "light-blue": 94,
    "light-magenta": 95,
    "light-cyan": 96,
}


def color_parts(color: Color, bold: bool = False) -> tuple[str, str]:
    if bold:
        return BOLD_COLOR_SEQ % COLOR_INDEX[color], RESET_SEQ
    return REG_COLOR_SEQ % COLOR_INDEX[color], RESET_SEQ


def uncolored(s: str) -> str:
    return re.sub(r"\033\[[\d;]+m", "", s)


def colored(s: str, color: Color | None = None, bold: bool = False) -> str:
    if color is None:
        return s
    start, end = color_parts(color, bold=bold)
    return start + s + end


def wrapped(
    s: str,
    length: int | None = None,
    space: str = " ",
    spaces: str | re.Pattern = r" ",
    newlines: str | re.Pattern = r"[\n\r]",
    too_long_suffix: str = "...",
    show_end: int | float | None = 0.5,
    max_lines: int | None = None,
) -> list[str]:
    strings = []
    lines = re.split(newlines, s.strip(), flags=re.MULTILINE | re.UNICODE)

    for line in lines:
        cur_string = []
        cur_length = 0

        for part in re.split(spaces, line.strip(), flags=re.MULTILINE | re.UNICODE):
            if length is None:
                cur_string.append(part)
                cur_length += len(space) + len(part)
            else:
                if len(part) > length:
                    if show_end is None:
                        part = part[: length - len(too_long_suffix)] + too_long_suffix
                    else:
                        if isinstance(show_end, float):
                            show_end = round(length * show_end)
                        part = (
                            part[: max(length - len(too_long_suffix) - show_end, 0)]
                            + too_long_suffix
                            + part[max(len(part) - show_end, 0) :]
                        )
                if cur_length + len(part) > length:
                    strings.append(space.join(cur_string))
                    cur_string = [part]
                    cur_length = len(part)
                else:
                    cur_string.append(part)
                    cur_length += len(space) + len(part)
        if cur_length > 0:
            strings.append(space.join(cur_string))

    if max_lines is not None:
        if max_lines < 1:
            raise ValueError("`max_lines` must be at least 1")
        if len(strings) > max_lines:
            strings = strings[-max_lines:]
            first_string, strings = strings[0], strings[1:]
            if length is not None and len(first_string) > length - len(too_long_suffix):
                first_string = first_string[len(too_long_suffix) :]
            strings = [too_long_suffix + first_string] + strings[: max_lines - 1]

    return strings


def outlined(
    s: str,
    inner: Color | None = None,
    side: Color | None = None,
    bold: bool = False,
    max_length: int | None = None,
    space: str = " ",
    spaces: str | re.Pattern = r" ",
    newlines: str | re.Pattern = r"[\n\r]",
) -> str:
    strs = wrapped(uncolored(s), max_length, space, spaces, newlines)
    max_len = max(len(s) for s in strs)
    strs = [f"{s}{' ' * (max_len - len(s))}" for s in strs]
    strs = [colored(s, inner, bold=bold) for s in strs]
    strs_with_sides = [f"{colored('│', side)} {s} {colored('│', side)}" for s in strs]
    top = colored("┌─" + "─" * max_len + "─┐", side)
    bottom = colored("└─" + "─" * max_len + "─┘", side)
    return "\n".join([top] + strs_with_sides + [bottom])


def show_info(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-cyan", side="cyan", bold=True)
    else:
        s = colored(s, "light-cyan", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


def show_error(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-red", side="red", bold=True)
    else:
        s = colored(s, "light-red", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


def show_warning(s: str, important: bool = False) -> None:
    if important:
        s = outlined(s, inner="light-yellow", side="yellow", bold=True)
    else:
        s = colored(s, "light-yellow", bold=False)
    sys.stdout.write(s)
    sys.stdout.write("\n")
    sys.stdout.flush()


class TextBlock:
    def __init__(
        self,
        text: str,
        color: Color | None = None,
        bold: bool = False,
        width: int | None = None,
        space: str = " ",
        spaces: str | re.Pattern = r" ",
        newlines: str | re.Pattern = r"[\n\r]",
        too_long_suffix: str = "...",
        no_sep: bool = False,
        center: bool = False,
        show_end: int | float | None = 0.5,
        max_lines: int | None = None,
    ) -> None:
        super().__init__()

        self.width = width
        self.lines = wrapped(uncolored(text), width, space, spaces, newlines, too_long_suffix, show_end, max_lines)
        self.color = color
        self.bold = bold
        self.no_sep = no_sep
        self.center = center


def render_text_blocks(
    blocks: list[list[TextBlock]],
    newline: str = "\n",
    align_all_blocks: bool = False,
    padding: int = 0,
) -> str:
    """Renders a collection of blocks into a single string.

    Args:
        blocks: The blocks to render.
        newline: The string to use as a newline separator.
        align_all_blocks: If set, aligns the widths for all blocks.
        padding: The amount of padding to add to each block.

    Returns:
        The rendered blocks.
    """
    if align_all_blocks:
        if any(len(row) != len(blocks[0]) for row in blocks):
            raise ValueError("All rows must have the same number of blocks in order to align them")
        widths = [[max(len(line) for line in i.lines) if i.width is None else i.width for i in r] for r in blocks]
        row_widths = [max(i) for i in zip(*widths)]
        for row in blocks:
            for i, block in enumerate(row):
                block.width = row_widths[i]

    def get_widths(row: list[TextBlock], n: int = 0) -> list[int]:
        return [
            (max(len(line) for line in block.lines) if block.width is None else block.width) + n + padding
            for block in row
        ]

    def get_acc_widths(row: list[TextBlock], n: int = 0) -> list[int]:
        return list(itertools.accumulate(get_widths(row, n)))

    def get_height(row: list[TextBlock]) -> int:
        return max(len(block.lines) for block in row)

    def pad(s: str, width: int, center: bool) -> str:
        swidth = len(s)
        if center:
            lpad, rpad = (width - swidth) // 2, (width - swidth + 1) // 2
        else:
            lpad, rpad = 0, width - swidth
        return " " * lpad + s + " " * rpad

    lines = []
    prev_row: list[TextBlock] | None = None
    for row in blocks:
        if prev_row is None:
            lines += ["┌─" + "─┬─".join(["─" * width for width in get_widths(row)]) + "─┐"]
        elif not all(block.no_sep for block in row):
            ins, outs = get_acc_widths(prev_row, 3), get_acc_widths(row, 3)
            segs = sorted([(i, False) for i in ins] + [(i, True) for i in outs])
            line = ["├"]

            c = 1
            for i, (s, is_out) in enumerate(segs):
                if i > 0 and segs[i - 1][0] == s:
                    continue
                is_in_out = i < len(segs) - 1 and segs[i + 1][0] == s
                is_last = i == len(segs) - 2 if is_in_out else i == len(segs) - 1

                line += "─" * (s - c)
                if is_last:
                    if is_in_out:
                        line += "┤"
                    elif is_out:
                        line += "┐"
                    else:
                        line += "┘"
                else:  # noqa: PLR5501
                    if is_in_out:
                        line += "┼"
                    elif is_out:
                        line += "┬"
                    else:
                        line += "┴"
                c = s + 1

            lines += ["".join(line)]

        for i in range(get_height(row)):
            lines += [
                "│ "
                + " │ ".join(
                    [
                        (
                            " " * width
                            if i >= len(block.lines)
                            else colored(pad(block.lines[i], width, block.center), block.color, bold=block.bold)
                        )
                        for block, width in zip(row, get_widths(row))
                    ]
                )
                + " │"
            ]

        prev_row = row
    if prev_row is not None:
        lines += ["└─" + "─┴─".join(["─" * width for width in get_widths(prev_row)]) + "─┘"]

    return newline.join(lines)


def format_timedelta(timedelta: datetime.timedelta, short: bool = False) -> str:
    """Formats a delta time to human-readable format.

    Args:
        timedelta: The delta to format
        short: If set, uses a shorter format

    Returns:
        The human-readable time delta
    """
    parts = []
    if timedelta.days > 0:
        if short:
            parts += [f"{timedelta.days}d"]
        else:
            parts += [f"{timedelta.days} day" if timedelta.days == 1 else f"{timedelta.days} days"]

    seconds = timedelta.seconds

    if seconds > 60 * 60:
        hours, seconds = seconds // (60 * 60), seconds % (60 * 60)
        if short:
            parts += [f"{hours}h"]
        else:
            parts += [f"{hours} hour" if hours == 1 else f"{hours} hours"]

    if seconds > 60:
        minutes, seconds = seconds // 60, seconds % 60
        if short:
            parts += [f"{minutes}m"]
        else:
            parts += [f"{minutes} minute" if minutes == 1 else f"{minutes} minutes"]

    if short:
        parts += [f"{seconds}s"]
    else:
        parts += [f"{seconds} second" if seconds == 1 else f"{seconds} seconds"]

    return ", ".join(parts)


class ColoredFormatter(logging.Formatter):
    """Defines a custom formatter for displaying logs."""

    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    COLORS: dict[str, Color] = {
        "WARNING": "yellow",
        "INFOALL": "magenta",
        "INFO": "cyan",
        "DEBUGALL": "grey",
        "DEBUG": "grey",
        "CRITICAL": "yellow",
        "FATAL": "red",
        "ERROR": "red",
    }

    def __init__(
        self,
        *,
        prefix: str | None = None,
        use_color: bool = True,
    ) -> None:
        asc_start, asc_end = color_parts("grey")
        name_start, name_end = color_parts("blue", bold=True)
        message_pre = [
            "{levelname:^19s}",
            asc_start,
            "{asctime}",
            asc_end,
            " [",
            name_start,
            "{name}",
            name_end,
            "]",
        ]
        message_post = [" {message}"]
        if prefix is not None:
            message_pre += [" ", colored(prefix, "magenta", bold=True)]
        message = "".join(message_pre + message_post)
        super().__init__(message, style="{", datefmt="%Y-%m-%d %H:%M:%S")

        self.use_color = use_color

    def format(self, record: logging.LogRecord) -> str:
        levelname = record.levelname

        match levelname:
            case "DEBUG":
                record.levelname = ""
            case "INFOALL":
                record.levelname = "INFO"
            case "DEBUGALL":
                record.levelname = "DEBUG"

        if record.levelname and self.use_color and levelname in self.COLORS:
            record.levelname = colored(record.levelname, self.COLORS[levelname], bold=True)
        return logging.Formatter.format(self, record)


def configure_logging(prefix: str | None = None) -> None:
    """Instantiates logging.

    Args:
        prefix: An optional prefix to add to the logger
    """
    root_logger = logging.getLogger()

    # Remove all existing handlers.
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    # Captures warnings from the warnings module.
    logging.captureWarnings(True)

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(ColoredFormatter(prefix=prefix))

    root_logger.addHandler(stream_handler)
    root_logger.setLevel(logging.INFO)



"""Defines a class for loading data from a dataset.

This is a greatly simplified dataloader implementation. We start a pool of
worker processes which continually load data from the dataset and put it on a
queue. The main process then reads from the queue and yields batches of data
as lists. The user is responsible for performing any additional downstream
tasks like collation.
"""

import logging
import multiprocessing as mp
import os
import random
import threading
import time
from dataclasses import dataclass
from multiprocessing.context import BaseContext
from queue import Queue
from threading import Event
from types import TracebackType
from typing import Callable, Generic, Self, TypeVar

import numpy as np

from dpshdl.dataset import Dataset, ErrorHandlingDataset
from dpshdl.testing import print_sample, run_test
from dpshdl.utils import configure_logging

logger = logging.getLogger(__name__)

T = TypeVar("T")
Tc = TypeVar("Tc")


class Timer:
    def __init__(self) -> None:
        super().__init__()

        self.start_time = 0.0
        self.elapsed_time = 0.0

    def __enter__(self) -> Self:
        self.start_time = time.time()
        return self

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        self.elapsed_time = time.time() - self.start_time


@dataclass(frozen=True)
class Stats:
    worker_id: int
    elapsed_time: float


@dataclass(frozen=True)
class DataloaderItem(Generic[T]):
    item: T | None
    exception: BaseException | None
    stats: Stats


@dataclass(frozen=True)
class CollatedDataloaderItem(Generic[T]):
    item: T | None
    exception: BaseException | None
    stats: list[Stats]


def init_random(worker_id: int = 0) -> None:
    random.seed(1337 + worker_id)
    np.random.seed(1337 + worker_id)


def dataloader_worker_init_fn(worker_id: int, num_workers: int) -> None:
    configure_logging(prefix=f"{worker_id}")
    init_random(worker_id)


def collate_worker_init_fn() -> None:
    configure_logging(prefix="col")
    init_random(-1)


def dataloader_worker(
    worker_init_fn: Callable[[int, int], None],
    dataset: Dataset[T, Tc],
    queue: "Queue[DataloaderItem[T]]",
    stop_event: Event,
    worker_id: int,
    num_workers: int,
    raise_errs: bool,
) -> None:
    worker_init_fn(worker_id, num_workers)
    dataset.worker_init(worker_id, num_workers)
    dataset_iterator = dataset.__iter__()
    while True:
        if stop_event.is_set():
            break
        try:
            with Timer() as timer:
                sample = dataset_iterator.__next__()
            queue.put(DataloaderItem(sample, None, Stats(worker_id, timer.elapsed_time)))
        except BaseException as e:
            if raise_errs:
                raise
            queue.put(DataloaderItem(None, e, Stats(worker_id, 0.0)))
            break


def collate_worker(
    worker_init_fn: Callable[[], None],
    samples_queue: "Queue[DataloaderItem[T]]",
    collated_queue: "Queue[CollatedDataloaderItem[Tc]]",
    collate_fn: Callable[[list[T]], Tc | None],
    stop_event: Event,
    batch_size: int,
    raise_errs: bool,
) -> None:
    worker_init_fn()
    samples: list[DataloaderItem[T]] = []
    while True:
        if stop_event.is_set():
            break
        sample = samples_queue.get()
        if sample.exception is not None:
            collated_queue.put(CollatedDataloaderItem(None, sample.exception, [sample.stats]))
            break
        if sample.item is None:
            raise RuntimeError("`item` should not be `None` unless there was an exception")
        samples.append(sample)
        if len(samples) == batch_size:
            try:
                if (collated := collate_fn([s.item for s in samples if s.item is not None])) is not None:
                    collated_queue.put(CollatedDataloaderItem(collated, None, [s.stats for s in samples]))
            except BaseException as e:
                if raise_errs:
                    raise
                collated_queue.put(CollatedDataloaderItem(None, e, [s.stats for s in samples]))
                break
            finally:
                samples = []


def default_num_workers(default: int) -> int:
    if hasattr(os, "sched_getaffinity"):
        try:
            return len(os.sched_getaffinity(0))
        except Exception:
            pass
    if (cpu_count := os.cpu_count()) is not None:
        return cpu_count
    return default


class Dataloader(Generic[T, Tc]):
    """Defines a dataloader for loading data from a dataset.

    Usage:

    .. code-block:: python

        with Dataloader(dataset, num_workers=4, batch_size=4) as dataloader:
            for batch in dataloader:
                # Do something with batch

    Parameters:
        dataset: The dataset to load data from.
        collate_fn: The function to use to collate the samples into a batch.
        num_workers: The number of workers to use. If 0, the dataset will be
            loaded on the main process. Otherwise, the dataset will be loaded
            on a pool of workers. If not provided, it will default to the
            number of CPUs on the system.
        batch_size: The batch size to use.
        prefetch_factor: The number of batches to pre-load from the dataset.
        ctx: The multiprocessing context to use. If not provided, the default
            context will be used.
        dataloader_worker_init_fn: The initialization function to use for
            the dataloader workers, which takes the workd ID and the number of
            workers as input.
        collate_worker_init_fn: The initialization function to use for the
            collate worker, which takes no inputs.
        item_callback: A callback function to call for each item before it is
            returned. This can be used for logging or debugging purposes.
        raise_errs: If set, raise worker errors instead of passing them to
            the error queue.
    """

    def __init__(
        self,
        dataset: Dataset[T, Tc],
        num_workers: int | None = None,
        batch_size: int = 1,
        prefetch_factor: int = 2,
        ctx: BaseContext | None = None,
        dataloader_worker_init_fn: Callable[[int, int], None] = dataloader_worker_init_fn,
        collate_worker_init_fn: Callable[[], None] = collate_worker_init_fn,
        item_callback: Callable[[CollatedDataloaderItem[Tc]], None] = lambda _: None,
        raise_errs: bool = False,
    ) -> None:
        super().__init__()

        if num_workers is None:
            num_workers = default_num_workers(0)
        if num_workers < 0:
            raise ValueError("`num_workers` must be greater than or equal to 0")
        if batch_size < 1:
            raise ValueError("`batch_size` must be greater than or equal to 1")
        if prefetch_factor < 1:
            raise ValueError("`prefetch_factor` must be greater than or equal to 1")

        self.dataset = dataset
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.ctx = mp.get_context() if ctx is None else ctx
        self.dataloader_worker_init_fn = dataloader_worker_init_fn
        self.collate_worker_init_fn = collate_worker_init_fn
        self.item_callback = item_callback
        self.raise_errs = raise_errs
        self.manager = self.ctx.Manager()

        self.processes: list[mp.Process] | None = None
        self.single_process_threads: tuple[threading.Thread, threading.Thread] | None = None
        self.samples_queue: Queue[DataloaderItem[T]] = self.manager.Queue(maxsize=batch_size * prefetch_factor)
        self.collated_queue: Queue[CollatedDataloaderItem[Tc]] = self.manager.Queue(maxsize=prefetch_factor)
        self.stop_event: Event = self.manager.Event()

    def test(
        self,
        max_samples: int = 10,
        log_interval: int | None = 1,
        print_fn: Callable[[int, Tc], None] = print_sample,
    ) -> None:
        """Defines a function for doing adhoc testing of the dataset.

        Args:
            max_samples: The maximum number of samples to test.
            log_interval: How often to log a sample. If None, don't log any
                samples.
            print_fn: The function to use for printing samples.
        """
        with self:
            run_test(
                ds=self,
                max_samples=max_samples,
                log_interval=log_interval,
                print_fn=print_fn,
            )

    def __iter__(self) -> "Dataloader[T, Tc]":
        return self

    def __next__(self) -> Tc:
        if self.processes is None and self.single_process_threads is None:
            raise RuntimeError("Dataloader is not running")
        item = self.collated_queue.get()
        self.item_callback(item)
        if item.exception:
            raise RuntimeError(f"Exception for worker ID(s) {[s.worker_id for s in item.stats]}: {item.exception}")
        if item.item is None:
            raise RuntimeError("`item` should not be `None` unless there was an exception")
        return item.item

    def __enter__(self) -> Self:
        if self.processes is None and self.single_process_threads is None:
            if self.num_workers > 0:
                self.processes = []
                for i in range(self.num_workers):
                    process = mp.Process(
                        target=dataloader_worker,
                        args=(
                            self.dataloader_worker_init_fn,
                            self.dataset,
                            self.samples_queue,
                            self.stop_event,
                            i,
                            self.num_workers,
                            self.raise_errs,
                        ),
                        daemon=True,
                        name=f"dataloader-worker-{i}",
                    )
                    process.start()
                    self.processes.append(process)
                collate_process = mp.Process(
                    target=collate_worker,
                    args=(
                        self.collate_worker_init_fn,
                        self.samples_queue,
                        self.collated_queue,
                        self.dataset.collate,
                        self.stop_event,
                        self.batch_size,
                        self.raise_errs,
                    ),
                    name="dataloader-worker-collate",
                )
                collate_process.start()
                self.processes.append(collate_process)
            else:
                dataloader_thread = threading.Thread(
                    target=dataloader_worker,
                    args=(
                        self.dataloader_worker_init_fn,
                        self.dataset,
                        self.samples_queue,
                        self.stop_event,
                        0,
                        1,
                        self.raise_errs,
                    ),
                )
                dataloader_thread.start()
                collate_thread = threading.Thread(
                    target=collate_worker,
                    args=(
                        self.collate_worker_init_fn,
                        self.samples_queue,
                        self.collated_queue,
                        self.dataset.collate,
                        self.stop_event,
                        self.batch_size,
                        self.raise_errs,
                    ),
                )
                collate_thread.start()
                self.single_process_threads = (dataloader_thread, collate_thread)
        else:
            raise RuntimeError("Dataloader is already running")
        return self

    def __exit__(self, _t: type[BaseException] | None, _e: BaseException | None, _tr: TracebackType | None) -> None:
        if self.processes is None and self.single_process_threads is None:
            raise RuntimeError("DataLoader is not running")

        # Stops the dataloader workers by setting the stop event and clearing
        # the queue. Clearing the queue is important because otherwise the
        # workers will block while trying to put items on the queue.
        self.stop_event.set()
        try:
            while not self.collated_queue.empty():
                self.collated_queue.get_nowait()
        except Exception:
            pass
        try:
            while not self.samples_queue.empty():
                self.samples_queue.get_nowait()
        except Exception:
            pass

        # Joins the dataloader workers.
        if self.processes is not None:
            for process in self.processes:
                process.terminate()
            self.processes = None
        elif self.single_process_threads is not None:
            dataloader_thread, collate_thread = self.single_process_threads
            dataloader_thread.join()
            collate_thread.join()
            self.single_process_thread = None
        else:
            raise RuntimeError("Unexpected state")


class _DummyDataset(Dataset[int, list[int]]):
    def next(self) -> int:
        return random.randint(0, 10)

    def collate(self, items: list[int]) -> list[int]:
        assert random.random() > 0.1, "A random error"
        time.sleep(random.random() * 0.1)
        return items


def test_error_handling_dataset_adhoc(test_samples: int = 1000000) -> None:
    """Tests the error handling dataset."""
    Dataloader(
        ErrorHandlingDataset(
            _DummyDataset(),
            flush_every_n_steps=test_samples,
            flush_every_n_seconds=None,
        ),
        batch_size=2,
    ).test(
        max_samples=test_samples,
        print_fn=lambda i, sample: logger.info("Sample %d: %s", i, sample),
    )


if __name__ == "__main__":
    # python -m dpshdl.dataloader
    test_error_handling_dataset_adhoc()
